
// rec_SMPTE_clock.js
// central time handler for render_node.
//
// randy jones
// rej@2uptech.com

// convert DSP time in fractional milliseconds to SMPTE at 29.97 fps
// don't worry about drop frames 

// destinations for outlet messages:
// recbangs smpte subframe dsptime render display looptime reclogic 

// version 0.6



post ("__SMPTE_clock\n");
autowatch = 1;

// global varables and code
inlets = 1;
outlets = 1;

var gFPS = 29.97;
var subframes_per_frame = 1;

var gPrerollLength = 2000.;
var gLoopStart = 0.;
var gLoopEnd = 10000.;
var gTotalLength = 60000.;
var gCountdownEnable = 0;
var gLoop = 1;
var gRecord = 0;
var gPreview = 1;
var gPluginCount = 0;

var outStr = new String();
var currFrame = 0;
var gSubFrame = 0;
var bang = "bang";
var gDspTime = 0.;
var gWorldTime = 0.;					// master time in ms
var gWorldTimeOffset = 0.;				// offset from dsp clock time.
var gPrevWorldTime = 0.
var gFrameTime = 0.;
var gShutterSpeed = 1./60.;


// countdown clock stuff
var mySketch = new JitterObject("jit.gl.sketch", "rn");
mySketch.position = [0., 0., 0.];
mySketch.automatic = 0;
mySketch.blend_enable = 0;
mySketch.lighting_enable = 0;	

var myText = new JitterObject("jit.gl.text3d", "rn");
myText.position = [0, 0, 0];
myText.align = 1;
myText.color = [1, 1, 1, 0.5];
myText.automatic = 0;
myText.blend_enable = 1;
myText.depth_enable = 0;

function renderCountdown(time)
{
	var t = Math.abs(time);
	var secStr;
	var timeStr = msecs_to_SMPTE(time);
 
 	outlet(0, "render", "erase");		
	outlet(0, "render", "ortho", 1);		
	outlet(0, "render", "camera", 0, 0, 2);		
	outlet(0, "render", "up", 0, 1, 0);		
	outlet(0, "render", "lookat", 0, 0, 0);		
	mySketch.reset();
	mySketch.shapeslice(360, 360);
	mySketch.gllinewidth(5);
	
	mySketch.depth_enable = 1;
	mySketch.glcolor(0, 0, 0, 1);
	mySketch.moveto(0, 0, 2);
	mySketch.quad(5, 5, 0, 5, -5, 0, -5, -5, 0, -5, 5, 0);	
	mySketch.draw();
  
  	if (time < -1000.)
  	{
		mySketch.depth_enable = 0;
		mySketch.glcolor(0.5, 0.5, 0.5, 1.);
		mySketch.moveto(0, 0, 1);
		mySketch.circle(0.45, 0., ((t % 1000.) / 1000. * 360.));
		mySketch.circle(0.45, 0., 4.);
		
		mySketch.glcolor(1., 1., 1., 1.);
		mySketch.moveto(0, 0, 1.1);
		mySketch.framecircle(0.5, 0., 360.);
		mySketch.framecircle(0.45, 0., ((t % 1000.) / 1000. * 360.));
		mySketch.framecircle(0.45, 0., 4.);
		
		mySketch.draw();
		
		myText.scale = [0.6, 0.6, 0.6];
		myText.position = [0.01, -0.2, 0.];
		myText.align = 1;
		myText.blend_mode = [6, 1];
		secStr = timeStr.slice(8, 9);	
		myText.text(secStr);
		myText.draw();
		
		myText.scale = [0.08, 0.08, 0.08];
		myText.position = [0.55, -0.75, 0.];
		myText.align = 0;
		myText.blend_mode = [6, 1];
		myText.text(timeStr);
		myText.draw();
	}
	
	outlet(0, "render", "ortho", 0);		
	
}



function renderClear()
{
	outlet(0, "render", "erase");		
	outlet(0, "render", "ortho", 1);		
	outlet(0, "render", "camera", 0, 0, 2);		
	outlet(0, "render", "up", 0, 1, 0);		
	outlet(0, "render", "lookat", 0, 0, 0);		
	mySketch.reset();
	
	mySketch.depth_enable = 0;
	mySketch.glcolor(0, 0, 0, 1);
	mySketch.moveto(0, 0, 2);
	mySketch.quad(5, 5, 0, 5, -5, 0, -5, -5, 0, -5, 5, 0);	
	mySketch.draw();
  
  	outlet(0, "render", "drawswap");	
	outlet(0, "render", "ortho", 0);		
	
}


// converts floats from dsptime~ to world time, handles record and loop logic, and ticks clock.
function msg_float(t)
{
	var was_before_loop_end, is_after_loop_end;
	var do_tick = 1;
	
	gDspTime = t;
	gPrevWorldTime = gWorldTime;
	gWorldTime = gDspTime + gWorldTimeOffset;
	
	was_before_loop_end = (gPrevWorldTime < gLoopEnd);
	is_after_loop_end = (gWorldTime > gLoopEnd);
	
	// handle loop
	if (was_before_loop_end && is_after_loop_end)
	{
		if (gLoop)
		{
			outlet(0, "phasor", 1000./gTotalLength, gLoopStart / gTotalLength);		// go, set phase
			gWorldTimeOffset = gLoopStart - gDspTime;
			gWorldTime = gLoopStart;
		}
		else
		{
			outlet(0, "phasor", 0., 0.);			// phasor stop
			outlet(0, "reclogic", "stop");
			do_tick = 0;
		}
	}
	
	if (gWorldTime > gTotalLength)
	{
		outlet(0, "phasor", 0., 0.);			// phasor stop
		outlet(0, "reclogic", "stop");
		do_tick = 0;
	}
	
	if (do_tick)
	{
		tick();
	}
}


var rec_subframes = 0;


// main clock routine
function tick()
{
	var totalframes, subframes, intframes, frame_interval;
	var f;
	
	totalsubframes = (gWorldTime * gFPS * subframes_per_frame * 0.001);		// total world time in subframes.	
	totalframes = totalsubframes / subframes_per_frame;
	subframes = Math.floor ((totalframes - Math.floor(totalframes)) * subframes_per_frame);
	frames = Math.floor(totalframes);// % gFPS;

		outStr = msecs_to_SMPTE(gWorldTime);
		outlet(0, "smpte", outStr);			// SMPTE time sent each frame.
		
		
	if (gCountdownEnable) // show render countdown, no subframe interpolation.
	{
		doCountdown()
	}	
	else 
	{
		outlet(0, "subframe", 0);		// clear

		// render
		for (f=0; f<subframes_per_frame; f++)	// always 1 in realtime
		{	
			gSubFrame = f + 1;
			subframe_interval = (1000.*gShutterSpeed) / subframes_per_frame;
			gFrameTime = gWorldTime + f*subframe_interval;

// post("t ", gWorldTime, "+", f, "*", subframe_interval, "=", gFrameTime, "\n");
			outlet(0, "worldtime", gFrameTime);	// world time outlet controls rendering patch

			// capture/record steps: 
			// update sim, 
			// render->erase, render->draw, render->to_texture, render->swap, 
			// bang texture.

			
			outlet(0, "render", "erase");
			
			if (gPluginCount == 0)
			{
				if (gRecord || gPreview)
				{
					outlet(0, "subframe", gSubFrame);		// subframe, 1 indexed
				}
				
				outlet(0, "render", "bang");	
				
				if (gRecord || gPreview)
				{
					outlet(0, "render", "to_texture");	
					if (gRecord)
					{
						outlet(0, "render", "swap");		
					}
					outlet(0, "accum", "jit_gl_texture", "rn_subframe");		
				}
			}
			else
			{
				outlet(0, "render", "drawclients");	
				outlet(0, "render", "to_texture");	
				outlet(0, "accum", "jit_gl_texture", "rn_subframe");		
				outlet(0, "subframe", gSubFrame);		// subframe, 1 indexed
				outlet(0, "videoplane", "bang");	
				outlet(0, "render", "swap");	
				outlet(0, "donebangs", "bang");	
				
			}
		}			
	} 
}



function doCountdown()
{
	var psecond, was_before_second, is_after_second;
	if (gWorldTime < gLoopStart)
	{
		gSubFrame = subframes_per_frame;		// force record
		outlet(0, "subframe", 1);				// clear accum
		outlet(0, "phasor", 0., 0.);			// phasor stop 
		if (gWorldTime > gLoopStart - gPrerollLength)
		{
			outlet(0, "render", "erase");

			renderCountdown(gWorldTime - gLoopStart);	


			if (gPluginCount == 0)
			{
				if (gRecord || gPreview)
				{
					outlet(0, "render", "to_texture");	
					outlet(0, "render", "bang");		
					outlet(0, "accum", "jit_gl_texture", "rn_subframe");		
				}
			}
			else
			{
				outlet(0, "render", "drawclients");	
				outlet(0, "render", "to_texture");	
				outlet(0, "accum", "jit_gl_texture", "rn_subframe");		
				outlet(0, "videoplane", "bang");	
				outlet(0, "render", "swap");	
				outlet(0, "donebangs", "bang");	
				
			}

			// preroll pip every second
			psecond = Math.floor((gLoopStart - gPrevWorldTime) * 0.001) * 1000.;
			was_before_second = (gLoopStart - gPrevWorldTime) > psecond;	// counting down
			is_after_second = (gLoopStart - gWorldTime) < psecond;
			if (was_before_second && is_after_second)
			{
				outlet(0, "preroll", "pip");	
			} 
		}
	}
	else
	{
		gCountdownEnable = 0;
		outlet(0, "phasor", 1000./gTotalLength, gLoopStart / gTotalLength);		// go, set phase
		outlet(0, "preroll", "mute", 0);
	}
}


// called from renderer's done bangs.  
function done()		
{
// post ("doned", gSubFrame, "of", subframes_per_frame, "\n");
	if (gRecord)
	{						
		if (gSubFrame == subframes_per_frame)
		{
// post ("recording frame", frames, "\n");
			outlet(0, "recbangs", bang);
		}
		else
		{
//	post ("nope\n");
		}
	}
}




// convert time in msecs to SMPTE at current frame rate.
function msecs_to_SMPTE(ftime)
{
	var totalframes, total, frames, seconds, minutes, hours;
	var outStr, negStr;
	var neg = (ftime < 0.);
	var t = ftime; 
	var intframes = Math.ceil(gFPS);
	
	if (neg)
	{
		t = -t;
		negStr = "-";
	}
	else
	{
		negStr = " ";
	}
		
	totalframes = (t * gFPS * 0.001);
	total = Math.floor(totalframes);
	
	frames = total % intframes;
	total = (total - frames) / intframes;
	
	seconds = total % 60;
	total = (total - seconds) / 60;
	
	minutes = total % 60;
	total = (total - minutes) / 60;
	
	hours = total % 24;
	total = (total - hours) / 24;
	
	outStr = negStr + D2(hours) + ":" + D2(minutes) + ":" + D2(seconds) + ":" + D2(frames);
	return(outStr);
}


function D2(n)
{
	var s = (((n) > 9 ? "" : "0") + n);
	return s.slice(0, 2);
}

 

// set dsp time without tick.
function dsptime(t)
{
	gDspTime = t;
}


// handle direct world time input
function worldtime(w)
{
	gWorldTime = w;
	gWorldTimeOffset = w - gDspTime;
}

function scrub(w)
{
	worldtime(w);
	gCountdownEnable = 0;
	outlet(0, "phasor", 1000./gTotalLength, gWorldTime / gTotalLength);	// go, set phase
	tick();
}


// start play from left loop point - countdown.
function play()
{
	gRecord = 0;
	currSubframe = -1;
	currFrame = 0;
	gDspTime = 0.;
	gWorldTimeOffset = gLoopStart - gPrerollLength;
	gCountdownEnable = 1;
	outlet(0, "preroll", "mute", 1);
}

// start record from left loop point - countdown.
function record()
{
	gRecord = 1;
	currSubframe = -1;
	currFrame = 0;
	gDspTime = 0.;
	gWorldTimeOffset = gLoopStart - gPrerollLength;
	gCountdownEnable = 1;
	outlet(0, "preroll", "mute", 1);
}


function preroll(n)
// set countdown in seconds
{
	gPrerollLength = n * 1000;
}


function length(n)
// set total length in ms, used for mandatory stop
{
	gTotalLength = n;
}

function loop_points(t1, t2)
{
	var s1, s2;
	
	gLoopStart = t1;
	gLoopEnd = t2;
	s1 = msecs_to_SMPTE(t1);	
	s2 = msecs_to_SMPTE(t2);
	outlet(0, "display", "loop", s1, s2);
}


function loop(m)
{
	gLoop = m;
}


function fps(n)
{
	gFPS = n;		
}

function subframes(n)
{
	subframes_per_frame = n;
}


function preview(n)
{
	gPreview = n;
}

function plugin_count(n)
{
	gPluginCount = n;
}

function shutter(n)
{
	gShutterSpeed = 1./(30.*(n+1));
}
